@@ -1,18 +1,23 @@
-from django.core.exceptions import ImproperlyConfigured
+from django.core.exceptions import ValidationError
 from django.core.validators import validate_email
 
-from allauth.account import app_settings as account_app_settings
+from allauth.account import app_settings as account_settings
 from allauth.account.adapter import get_adapter as get_account_adapter
+from allauth.account.fields import PhoneField
 from allauth.account.forms import (
     AddEmailForm,
     BaseSignupForm,
+    ChangeEmailForm,
+    ChangePhoneForm,
     ConfirmLoginCodeForm,
     ReauthenticateForm,
     RequestLoginCodeForm,
     ResetPasswordForm,
     UserTokenForm,
+    VerifyPhoneForm,
 )
 from allauth.account.internal import flows
+from allauth.account.internal.textkit import compare_code
 from allauth.account.models import (
     EmailAddress,
     Login,
@@ -26,6 +31,14 @@ from allauth.headless.internal.restkit import inputs
 class SignupInput(BaseSignupForm, inputs.Input):
     password = inputs.CharField()
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        password = account_settings.SIGNUP_FIELDS.get("password1")
+        if not password:
+            del self.fields["password"]
+        else:
+            self.fields["password"].required = password["required"]
+
     def clean_password(self):
         password = self.cleaned_data["password"]
         return get_account_adapter().clean_password(password)
@@ -34,50 +47,33 @@ class SignupInput(BaseSignupForm, inputs.Input):
 class LoginInput(inputs.Input):
     username = inputs.CharField(required=False)
     email = inputs.EmailField(required=False)
+    # NOTE: Always require E164, no need to use adapter.phone_form_field
+    phone = PhoneField(required=False)
     password = inputs.CharField()
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        for field in ["username", "email", "phone"]:
+            if field not in account_settings.LOGIN_METHODS:
+                del self.fields[field]
+        if len(account_settings.LOGIN_METHODS) == 1:
+            self.fields[next(iter(account_settings.LOGIN_METHODS))].required = True
+
     def clean(self):
         cleaned_data = super().clean()
-        username = None
-        email = None
-
-        if account_app_settings.LOGIN_METHODS == {
-            account_app_settings.LoginMethod.USERNAME
-        }:
-            username = cleaned_data.get("username")
-            missing_field = "username"
-        elif account_app_settings.LOGIN_METHODS == {
-            account_app_settings.LoginMethod.EMAIL
-        }:
-            email = cleaned_data.get("email")
-            missing_field = "email"
-        elif account_app_settings.LOGIN_METHODS == {
-            account_app_settings.LoginMethod.USERNAME,
-            account_app_settings.LoginMethod.EMAIL,
-        }:
-            username = cleaned_data.get("username")
-            email = cleaned_data.get("email")
-            missing_field = "email"
-            if email and username:
-                raise get_adapter().validation_error("email_or_username")
-        else:
-            raise ImproperlyConfigured(account_app_settings.LOGIN_METHODS)
-
-        if not email and not username:
-            if not self.errors.get(missing_field):
-                self.add_error(
-                    missing_field, get_adapter().validation_error("required")
-                )
-
+        if self.errors:
+            return cleaned_data
+        credentials = {}
+        for login_method in account_settings.LOGIN_METHODS:
+            value = cleaned_data.get(login_method)
+            if value is not None and login_method in self.data.keys():
+                credentials[login_method] = value
+        if len(credentials) != 1:
+            raise get_account_adapter().validation_error("invalid_login")
         password = cleaned_data.get("password")
-        if password and (username or email):
-            credentials = {"password": password}
-            if email:
-                credentials["email"] = email
-                auth_method = account_app_settings.AuthenticationMethod.EMAIL
-            else:
-                credentials["username"] = username
-                auth_method = account_app_settings.AuthenticationMethod.USERNAME
+        if password:
+            auth_method = next(iter(credentials.keys()))
+            credentials["password"] = password
             user = get_account_adapter().authenticate(context.request, **credentials)
             if user:
                 self.login = Login(user=user, email=credentials.get("email"))
@@ -96,20 +92,32 @@ class LoginInput(inputs.Input):
 class VerifyEmailInput(inputs.Input):
     key = inputs.CharField()
 
+    def __init__(self, *args, **kwargs):
+        self.process = kwargs.pop("process", None)
+        super().__init__(*args, **kwargs)
+
     def clean_key(self):
         key = self.cleaned_data["key"]
-        model = get_emailconfirmation_model()
-        confirmation = model.from_key(key)
-        valid = confirmation and not confirmation.key_expired()
-        if not valid:
-            raise get_account_adapter().validation_error(
-                "incorrect_code"
-                if account_app_settings.EMAIL_VERIFICATION_BY_CODE_ENABLED
-                else "invalid_or_expired_key"
-            )
-        if valid and not confirmation.email_address.can_set_verified():
+        if self.process:
+            if not compare_code(actual=key, expected=self.process.code):
+                raise get_account_adapter().validation_error("incorrect_code")
+            valid = True
+            email_address = self.process.email_address
+        else:
+            model = get_emailconfirmation_model()
+            verification = model.from_key(key)
+            valid = verification and not verification.key_expired()
+            if not valid:
+                raise get_account_adapter().validation_error(
+                    "incorrect_code"
+                    if account_settings.EMAIL_VERIFICATION_BY_CODE_ENABLED
+                    else "invalid_or_expired_key"
+                )
+            email_address = verification.email_address
+            self.verification = verification
+        if valid and not email_address.can_set_verified():
             raise get_account_adapter().validation_error("email_taken")
-        return confirmation
+        return key
 
 
 class RequestPasswordResetInput(ResetPasswordForm, inputs.Input):
@@ -120,10 +128,23 @@ class ResetPasswordKeyInput(inputs.Input):
     key = inputs.CharField()
 
     def __init__(self, *args, **kwargs):
-        self.user = None
+        self.user = kwargs.pop("user", None)
+        self.code = kwargs.pop("code", None)
         super().__init__(*args, **kwargs)
 
     def clean_key(self):
+        if account_settings.PASSWORD_RESET_BY_CODE_ENABLED:
+            return self._clean_key_code()
+        else:
+            return self._clean_key_link()
+
+    def _clean_key_code(self):
+        key = self.cleaned_data["key"]
+        if not compare_code(actual=key, expected=self.code):
+            raise get_account_adapter().validation_error("incorrect_code")
+        return key
+
+    def _clean_key_link(self):
         key = self.cleaned_data["key"]
         uidb36, _, subkey = key.partition("-")
         token_form = UserTokenForm(data={"uidb36": uidb36, "key": subkey})
@@ -140,7 +161,10 @@ class ResetPasswordInput(ResetPasswordKeyInput):
         cleaned_data = super().clean()
         password = self.cleaned_data.get("password")
         if self.user and password is not None:
-            get_account_adapter().clean_password(password, user=self.user)
+            try:
+                get_account_adapter().clean_password(password, user=self.user)
+            except ValidationError as e:
+                self.add_error("password", e)
         return cleaned_data
 
 
@@ -214,3 +238,15 @@ class RequestLoginCodeInput(RequestLoginCodeForm, inputs.Input):
 
 class ConfirmLoginCodeInput(ConfirmLoginCodeForm, inputs.Input):
     pass
+
+
+class VerifyPhoneInput(VerifyPhoneForm, inputs.Input):
+    pass
+
+
+class ChangePhoneInput(ChangePhoneForm, inputs.Input):
+    pass
+
+
+class ChangeEmailInput(ChangeEmailForm, inputs.Input):
+    pass
